#version 300 es
precision highp float;
precision highp isampler2D;
precision highp sampler2D;
precision highp sampler2DArray;

// Based on: http://wdobbie.com/post/gpu-text-rendering-with-vector-textures/

struct Glyph {
	int start, count;
};

struct Curve {
	vec2 p0, p1, p2;
};

// CURVE_SIZE
// is the amount of vec2's in the curve
// we can hardcode this, as we can simply
// count the floats in the struct
// 6 floats => 3 vec2's
#define CURVE_SIZE 3
const int CURVES_BUFFER_SIZE = {{CURVES_BUFFER_SIZE}};

layout(std140) uniform curves_t {
    vec4 data[{{CURVES_BUFFER_SIZE}}];
} curves;

uniform isampler2D glyphs;
ivec3 curvesTextureSize;
int curveBreakPoint;
//uniform sampler2D iChannel0;
//uniform vec4 color;

// Controls for debugging and exploring:

// Size of the window (in pixels) used for 1-dimensional anti-aliasing along each rays.
//   0 - no anti-aliasing
//   1 - normal anti-aliasing
// >=2 - exaggerated effect
uniform float antiAliasingWindowSize;

// Enable a second ray along the y-axis to achieve 2-dimensional anti-aliasing.
uniform bool enableSuperSamplingAntiAliasing;

// Draw control points for debugging (green - on curve, magenta - off curve).
uniform bool enableControlPointsVisualization;


in vec2 uv;
flat in int bufferIndex;
in vec4 color;

out vec4 result;

Glyph loadGlyph(int index) {
    Glyph result;
    ivec2 data = texelFetch(glyphs, ivec2(index, 0), 0).xy;
    result.start = data.x;
    result.count = data.y;
    return result;
}

Curve loadCurve(int index) {
    int index_a = int(floor((float(index) / 4.0) * 6.0));
    int index_b = index_a + 1;
    vec4 a = curves.data[index_a];
    vec4 b = curves.data[index_b];
    Curve curve;
    if (index % 2 == 0) {
        curve.p0 = vec2(a[0], a[1]);
        curve.p1 = vec2(a[2], a[3]);
        curve.p2 = vec2(b[0], b[1]);
    } else {
        curve.p0 = vec2(a[2], a[3]);
        curve.p1 = vec2(b[0], b[1]);
        curve.p2 = vec2(b[2], b[3]);
    }
    return curve;
}

float computeCoverage(float inverseDiameter, vec2 p0, vec2 p1, vec2 p2) {
	if (p0.y > 0.0 && p1.y > 0.0 && p2.y > 0.0) return 0.0;
	if (p0.y < 0.0 && p1.y < 0.0 && p2.y < 0.0) return 0.0;

	// Note: Simplified from abc formula by extracting a factor of (-2) from b.
	vec2 a = p0 - 2.0*p1 + p2;
	vec2 b = p0 - p1;
	vec2 c = p0;

	float t0, t1;
	if (abs(a.y) >= 1e-5) {
		// Quadratic segment, solve abc formula to find roots.
		float radicand = b.y*b.y - a.y*c.y;
		if (radicand <= 0.0) return 0.0;

		float s = sqrt(radicand);
		t0 = (b.y - s) / a.y;
		t1 = (b.y + s) / a.y;
	} else {
		// Linear segment, avoid division by a.y, which is near zero.
		// There is only one root, so we have to decide which variable to
		// assign it to based on the direction of the segment, to ensure that
		// the ray always exits the shape at t0 and enters at t1. For a
		// quadratic segment this works 'automatically', see readme.
		float t = p0.y / (p0.y - p2.y);
		if (p0.y < p2.y) {
			t0 = -1.0;
			t1 = t;
		} else {
			t0 = t;
			t1 = -1.0;
		}
	}

	float alpha = 0.0;

	if (t0 >= 0.0 && t0 < 1.0) {
		float x = (a.x*t0 - 2.0*b.x)*t0 + c.x;
		alpha += clamp(x * inverseDiameter + 0.5, 0.0, 1.0);
	}

	if (t1 >= 0.0 && t1 < 1.0) {
		float x = (a.x*t1 - 2.0*b.x)*t1 + c.x;
		alpha -= clamp(x * inverseDiameter + 0.5, 0.0, 1.0);
	}

	return alpha;
}

vec2 rotate(vec2 v) {
	return vec2(v.y, -v.x);
}

void main() {

    float alpha = 0.0;

    // Inverse of the diameter of a pixel in uv units for anti-aliasing.
    vec2 inverseDiameter = 1.0 / (antiAliasingWindowSize * fwidth(uv));

    vec4 c = color;

    Glyph glyph = loadGlyph(bufferIndex);
    for (int i = 0; i < glyph.count; i++) {
        //if (glyph.start + i > curveBreakPoint) {
            //c = vec4(1.0,0.0,0.0,1.0);
        //}
        Curve curve = loadCurve(glyph.start + i);

        vec2 p0 = curve.p0 - uv;
        vec2 p1 = curve.p1 - uv;
        vec2 p2 = curve.p2 - uv;

        alpha += computeCoverage(inverseDiameter.x, p0, p1, p2);
        if (enableSuperSamplingAntiAliasing) {
            alpha += computeCoverage(inverseDiameter.y, rotate(p0), rotate(p1), rotate(p2));
        }
    }

    if (enableSuperSamplingAntiAliasing) {
        alpha *= 0.5;
    }

    alpha = clamp(alpha, 0.0, 1.0);
    result = vec4(c.rgb, c.a * alpha);

    if (enableControlPointsVisualization) {
        // Visualize control points.
        vec2 fw = fwidth(uv);
        float r = 4.0 * 0.5 * (fw.x + fw.y);
        for (int i = 0; i < glyph.count; i++) {
            Curve curve = loadCurve(glyph.start + i);

            vec2 p0 = curve.p0 - uv;
            vec2 p1 = curve.p1 - uv;
            vec2 p2 = curve.p2 - uv;

            if (dot(p0, p0) < r*r || dot(p2, p2) < r*r) {
                result = vec4(0.0, 1.0, 0.0, 1.0);
                return;
            }

            if (dot(p1, p1) < r*r) {
                result = vec4(1.0, 0.0, 1.0, 1.0);
                return;
            }
        }
    }
}
